/*
  AVI2YUV.C  version 0.97
  
  Copyright (c) 1996/99 by Bernhard Schwall (b.schwall@gmx.de)

  Program to decode a MJPEG or uncompressed RGB AVI into single YUV pictures
  handled by the Berkeley MPEG encoder.

  Based on SHOWRIFF.C
  (c)94 UP-Vision Computergrafik for c't
  
  Some parts of the program are taken for libjpeg version 6.
  "this software is based in part on the word of the independant JPEG Group"
  
  This program is free software; you can redistribute it and/or modify
  it under the terms of the GNU General Public License as published by
  the Free Software Foundation;

  This program is distributed in the hope that it will be useful, but
  WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
  General Public License for more details.

  You should have received a copy of the GNU General Public License
  along with this program; if not, write to the Free Software
  Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
  
  version 0.1 : 31.01.1996
    first release
  version 0.2 : 14.02.1996
    possibility to save single frames as YUV
    typedefs moved to avi2yuv.h
    support of uncompressed 16 bit hicolor AVI
  version 0.3 : 01.03.1996
    fixed a bug which caused the output to be corrupt be certain horizontal
      resolutions.
    possibility to crop the video
    support of uncompressed 24 and 32 bit true color AVI
  version 0.4 : 23.04.96
    scaling routine changed
  version 0.5 : 26.06.96
    possibility to scale the picture up (will be centered)
  version 0.6 : 22.07.1996
    decoding of YVU9 AVIs (Indeo RAW)
  version 0.61: 03.09.1997
    print picture scale/rate
  version 0.7 : 07.11.1997
    decoding of Y41P (Haupauge WinTV format)
  version 0.8 : 03.02.1998
    possibility to save single frames as Windows BMP file
    JPEG decompression speed up
  version 0.81: 17.02.1998
    decoding of dmb1 AVIs (Matrox Rainbow Runner Studio)
    possibility to disable the prevention of the aspect ratio when scaling
  version 0.82: 21.03.1998
    fixed a problem with corrupt header by RGB AVIs
  version 0.9 : 06.04.1998
    possibility to save sound as wav file
    crop sound if start or endframes are skipped
  version 0.91: 28.10.1998
    fixed an error with framerate adjustment (-f)
  version 0.92: 23.12.1998
    fixed an error with corrupt JPEG frames inside the AVI
      (looks like an error of the RainbowRunner driver)
    added option -t to test the correctness of each frame without writing
      the output to stdout or file (only for MJPEG).
  version 0.93: 04.01.1999
    changed option -v to only show the header information
  version 0.94: 16.05.1999
    added possibility to save interlaced frames either as frams A+B
      or combined
  version 0.95: 17.05.1999
    save field of interleaved video in double Y resolution and
      with different numbers
  version 0.96: 30.05.1999
    FrameType changed from enum to typedef for easier handling
    don't stop decoding if filesize in main header is wrong
  version 0.97: 16.05.1999
    fixed error report for testing only without error found
*/

#include <stdlib.h>
#include <string.h>
#include <stdio.h>
#include <jpeglib.h>
#include <jinclude.h>
#include "avi2yuv.h"
#include <setjmp.h>

DWORD   LastChunkTyp = 0;
DWORD   PicNumber = 0;
DWORD   DroppedFrames = 0;
DWORD   CorruptFrames = 0;
uCHAR   *outBuffer;
uCHAR   *YBuffer, *UBuffer, *VBuffer;
boolean PrintFormat;
FILE    *WaveFile;
int     FrameSizeX, FrameSizeY;
int     FirstFrame = 1, LastFrame = -1;
int	CropX1 = 0, CropX2 = -1, CropY1 = 0, CropY2 = -1;
DWORD   MSPerFrame = 0;
DWORD   ActualMS, AVI_MS;
char    PicName[255];
boolean SaveFrames;
boolean KeepAspect;
boolean TestFrames;
int     PicFormat;
int         WaveFormat = 0;
tWaveHeader WaveHeader;
tGlobalInfo GlobalInfo;
long        SoundStart, SoundEnd, SoundPos;
FrameType   ftSave;
int         CountSkip = 0;
boolean     DoubleLines = FALSE;


/*******************************************/
/* Build a string from a FOURCC number     */
/* (s must have room for at least 5 chars) */
/*******************************************/

void FOURCC2Str(FOURCC fcc, char* s)
{
  s[0]=(fcc      ) & 0xFF;
  s[1]=(fcc >>  8) & 0xFF;
  s[2]=(fcc >> 16) & 0xFF;
  s[3]=(fcc >> 24) & 0xFF;
  s[4]=0;
}


/***************************************************************/
/* Reads a chunk ID and the chunk's size from file f at actual */
/* file position                                               */
/***************************************************************/

boolean ReadChunkHead(FILE* f, FOURCC* ID, DWORD* size)
{
  if (!fread(ID,sizeof(FOURCC),1,f)) return(FALSE);
  if (!fread(size,sizeof(DWORD),1,f)) return(FALSE);
  return(TRUE);
}


/*********************************************************************/
/* create the standard huffman table as the MJPG doesn't include one */
/* taken from jcparam.c (libjpeg 6)                                  */
/*********************************************************************/

void add_huff_table (j_decompress_ptr dinfo,
		JHUFF_TBL **htblptr, const UINT8 *bits, const UINT8 *val)
/* Define a Huffman table */
{
  if (*htblptr == NULL)
    *htblptr = jpeg_alloc_huff_table((j_common_ptr) dinfo);

  MEMCOPY((*htblptr)->bits, bits, SIZEOF((*htblptr)->bits));
  MEMCOPY((*htblptr)->huffval, val, SIZEOF((*htblptr)->huffval));

  /* Initialize sent_table FALSE so table will be written to JPEG file. */
  (*htblptr)->sent_table = FALSE;
}

void std_huff_tables (j_decompress_ptr dinfo)
/* Set up the standard Huffman tables (cf. JPEG standard section K.3) */
/* IMPORTANT: these are only valid for 8-bit data precision! */
{
  static const UINT8 bits_dc_luminance[17] =
    { /* 0-base */ 0, 0, 1, 5, 1, 1, 1, 1, 1, 1, 0, 0, 0, 0, 0, 0, 0 };
  static const UINT8 val_dc_luminance[] =
    { 0, 1, 2, 3, 4, 5, 6, 7, 8, 9, 10, 11 };
  
  static const UINT8 bits_dc_chrominance[17] =
    { /* 0-base */ 0, 0, 3, 1, 1, 1, 1, 1, 1, 1, 1, 1, 0, 0, 0, 0, 0 };
  static const UINT8 val_dc_chrominance[] =
    { 0, 1, 2, 3, 4, 5, 6, 7, 8, 9, 10, 11 };
  
  static const UINT8 bits_ac_luminance[17] =
    { /* 0-base */ 0, 0, 2, 1, 3, 3, 2, 4, 3, 5, 5, 4, 4, 0, 0, 1, 0x7d };
  static const UINT8 val_ac_luminance[] =
    { 0x01, 0x02, 0x03, 0x00, 0x04, 0x11, 0x05, 0x12,
      0x21, 0x31, 0x41, 0x06, 0x13, 0x51, 0x61, 0x07,
      0x22, 0x71, 0x14, 0x32, 0x81, 0x91, 0xa1, 0x08,
      0x23, 0x42, 0xb1, 0xc1, 0x15, 0x52, 0xd1, 0xf0,
      0x24, 0x33, 0x62, 0x72, 0x82, 0x09, 0x0a, 0x16,
      0x17, 0x18, 0x19, 0x1a, 0x25, 0x26, 0x27, 0x28,
      0x29, 0x2a, 0x34, 0x35, 0x36, 0x37, 0x38, 0x39,
      0x3a, 0x43, 0x44, 0x45, 0x46, 0x47, 0x48, 0x49,
      0x4a, 0x53, 0x54, 0x55, 0x56, 0x57, 0x58, 0x59,
      0x5a, 0x63, 0x64, 0x65, 0x66, 0x67, 0x68, 0x69,
      0x6a, 0x73, 0x74, 0x75, 0x76, 0x77, 0x78, 0x79,
      0x7a, 0x83, 0x84, 0x85, 0x86, 0x87, 0x88, 0x89,
      0x8a, 0x92, 0x93, 0x94, 0x95, 0x96, 0x97, 0x98,
      0x99, 0x9a, 0xa2, 0xa3, 0xa4, 0xa5, 0xa6, 0xa7,
      0xa8, 0xa9, 0xaa, 0xb2, 0xb3, 0xb4, 0xb5, 0xb6,
      0xb7, 0xb8, 0xb9, 0xba, 0xc2, 0xc3, 0xc4, 0xc5,
      0xc6, 0xc7, 0xc8, 0xc9, 0xca, 0xd2, 0xd3, 0xd4,
      0xd5, 0xd6, 0xd7, 0xd8, 0xd9, 0xda, 0xe1, 0xe2,
      0xe3, 0xe4, 0xe5, 0xe6, 0xe7, 0xe8, 0xe9, 0xea,
      0xf1, 0xf2, 0xf3, 0xf4, 0xf5, 0xf6, 0xf7, 0xf8,
      0xf9, 0xfa };
  
  static const UINT8 bits_ac_chrominance[17] =
    { /* 0-base */ 0, 0, 2, 1, 2, 4, 4, 3, 4, 7, 5, 4, 4, 0, 1, 2, 0x77 };
  static const UINT8 val_ac_chrominance[] =
    { 0x00, 0x01, 0x02, 0x03, 0x11, 0x04, 0x05, 0x21,
      0x31, 0x06, 0x12, 0x41, 0x51, 0x07, 0x61, 0x71,
      0x13, 0x22, 0x32, 0x81, 0x08, 0x14, 0x42, 0x91,
      0xa1, 0xb1, 0xc1, 0x09, 0x23, 0x33, 0x52, 0xf0,
      0x15, 0x62, 0x72, 0xd1, 0x0a, 0x16, 0x24, 0x34,
      0xe1, 0x25, 0xf1, 0x17, 0x18, 0x19, 0x1a, 0x26,
      0x27, 0x28, 0x29, 0x2a, 0x35, 0x36, 0x37, 0x38,
      0x39, 0x3a, 0x43, 0x44, 0x45, 0x46, 0x47, 0x48,
      0x49, 0x4a, 0x53, 0x54, 0x55, 0x56, 0x57, 0x58,
      0x59, 0x5a, 0x63, 0x64, 0x65, 0x66, 0x67, 0x68,
      0x69, 0x6a, 0x73, 0x74, 0x75, 0x76, 0x77, 0x78,
      0x79, 0x7a, 0x82, 0x83, 0x84, 0x85, 0x86, 0x87,
      0x88, 0x89, 0x8a, 0x92, 0x93, 0x94, 0x95, 0x96,
      0x97, 0x98, 0x99, 0x9a, 0xa2, 0xa3, 0xa4, 0xa5,
      0xa6, 0xa7, 0xa8, 0xa9, 0xaa, 0xb2, 0xb3, 0xb4,
      0xb5, 0xb6, 0xb7, 0xb8, 0xb9, 0xba, 0xc2, 0xc3,
      0xc4, 0xc5, 0xc6, 0xc7, 0xc8, 0xc9, 0xca, 0xd2,
      0xd3, 0xd4, 0xd5, 0xd6, 0xd7, 0xd8, 0xd9, 0xda,
      0xe2, 0xe3, 0xe4, 0xe5, 0xe6, 0xe7, 0xe8, 0xe9,
      0xea, 0xf2, 0xf3, 0xf4, 0xf5, 0xf6, 0xf7, 0xf8,
      0xf9, 0xfa };
  
  add_huff_table(dinfo, &dinfo->dc_huff_tbl_ptrs[0],
		 bits_dc_luminance, val_dc_luminance);
  add_huff_table(dinfo, &dinfo->ac_huff_tbl_ptrs[0],
		 bits_ac_luminance, val_ac_luminance);
  add_huff_table(dinfo, &dinfo->dc_huff_tbl_ptrs[1],
		 bits_dc_chrominance, val_dc_chrominance);
  add_huff_table(dinfo, &dinfo->ac_huff_tbl_ptrs[1],
		 bits_ac_chrominance, val_ac_chrominance);
}

/**************************************************************/
/* define a new error routine to handle corrupt JPEGs without */
/* program termination                                        */
/**************************************************************/

struct my_error_mgr {
  struct  jpeg_error_mgr pub;
  jmp_buf setjmp_buffer;
};

typedef struct my_error_mgr* my_error_ptr;

void my_error_exit(j_common_ptr cinfo) {
  my_error_ptr myerr = (my_error_ptr)cinfo->err;
  longjmp(myerr->setjmp_buffer, 1);
}

/*****************************************************************/
/* read the JPEG to outBuffer with format YUV (3 byte per pixel) */
/*****************************************************************/

int Decode_JPEG(FILE *inFile, char *outBuffer){
int row_stride;
int y, ret;
struct jpeg_decompress_struct dinfo;
struct my_error_mgr jerr;

  dinfo.err = jpeg_std_error(&jerr.pub);
  jerr.pub.error_exit = my_error_exit;
  if (setjmp(jerr.setjmp_buffer)) {
    jpeg_destroy_decompress(&dinfo);
    return 0;
  }
  jpeg_create_decompress(&dinfo);
  jpeg_stdio_src(&dinfo, inFile);
  ret = jpeg_read_header(&dinfo, FALSE);
  if (ret != JPEG_HEADER_OK) {
    fprintf(stderr, "header corrupt\n");
    CorruptFrames++;
    return 0;
  }
  if (dinfo.dc_huff_tbl_ptrs[0] == NULL){
    std_huff_tables(&dinfo);
  }
  dinfo.out_color_space = JCS_YCbCr;
  /* not as acurate as the original one but much faster */
  dinfo.dct_method = JDCT_IFAST;
  jpeg_start_decompress(&dinfo);
  row_stride= dinfo.output_width*dinfo.output_components;
  for (y = 0; y < dinfo.output_height; y++){
    jpeg_read_scanlines(&dinfo, (JSAMPARRAY) &outBuffer, 1);
    outBuffer += row_stride;
  }
  jpeg_destroy_decompress(&dinfo);
  return 1;
}


/*****************************************************************/
/* read two JPEG to outBuffer with format YUV (3 byte per pixel) */
/* this combines two interleaved frames                          */
/* DoubleLines MUST only be set with FrameA of FrameB            */
/*****************************************************************/

int Decode_JPEG_interlaced(FILE *inFile, char *outBuffer, 
                           FrameType ftDecode, boolean DoubleLines){
int row_stride;
int y, ret;
struct jpeg_decompress_struct dinfo;
struct my_error_mgr jerr;
unsigned long filepos;
unsigned int  buf[4];
char          *outBufTmp;

  filepos = ftell(inFile);
  // read the JPEG length
  fread(buf, 16, 1, inFile);
  // reposition the file to beginning of JPEG
  fseek(inFile, filepos, SEEK_SET);
  
  outBufTmp = outBuffer;

  if (ftDecode != FrameB)
  {  
    dinfo.err = jpeg_std_error(&jerr.pub);
    jerr.pub.error_exit = my_error_exit;
    if (setjmp(jerr.setjmp_buffer)) {
      jpeg_destroy_decompress(&dinfo);
      return 0;
    }
    jpeg_create_decompress(&dinfo);
    jpeg_stdio_src(&dinfo, inFile);
    ret = jpeg_read_header(&dinfo, FALSE);
    if (ret != JPEG_HEADER_OK) {
      fprintf(stderr, "header corrupt\n");
      CorruptFrames++;
      return 0;
    }
    if (dinfo.dc_huff_tbl_ptrs[0] == NULL){
      std_huff_tables(&dinfo);
    }
    dinfo.out_color_space = JCS_YCbCr;
    /* not as acurate as the original one but much faster */
    dinfo.dct_method = JDCT_IFAST;
    jpeg_start_decompress(&dinfo);
    row_stride= dinfo.output_width*dinfo.output_components;
    // decode interlaced
    if (ftDecode == FullFrame)
    {
      outBuffer += row_stride;
      row_stride *= 2;
    }
    for (y = 0; y < dinfo.output_height; y++){
      jpeg_read_scanlines(&dinfo, (JSAMPARRAY) &outBuffer, 1);
      if (DoubleLines == TRUE)
      {
        memcpy(outBuffer+row_stride, outBuffer, row_stride);
        outBuffer += row_stride;
      }
      outBuffer += row_stride;
    }
    jpeg_destroy_decompress(&dinfo);
  }
  
  if (ftDecode != FrameA)
  {
    // read the second frame
    outBuffer = outBufTmp;
    
    // convert MSB -> LSB
    buf[0] = (buf[3] >> 24) | ((buf[3] & 0x00ff0000) >> 8) | ((buf[3] & 0x0000ff00) << 8) |
             ((buf[3] & 0xff) << 24);
    fseek(inFile, filepos+buf[0], SEEK_SET);
    
    dinfo.err = jpeg_std_error(&jerr.pub);
    jerr.pub.error_exit = my_error_exit;
    if (setjmp(jerr.setjmp_buffer)) {
      jpeg_destroy_decompress(&dinfo);
      return 0;
    }
    jpeg_create_decompress(&dinfo);
    jpeg_stdio_src(&dinfo, inFile);
    ret = jpeg_read_header(&dinfo, FALSE);
    if (ret != JPEG_HEADER_OK) {
      fprintf(stderr, "header corrupt\n");
      CorruptFrames++;
      return 0;
    }
    if (dinfo.dc_huff_tbl_ptrs[0] == NULL){
      std_huff_tables(&dinfo);
    }
    dinfo.out_color_space = JCS_YCbCr;
    /* not as acurate as the original one but much faster */
    dinfo.dct_method = JDCT_IFAST;
    jpeg_start_decompress(&dinfo);
    row_stride= dinfo.output_width*dinfo.output_components;
    // decode interlaced
    if (ftDecode == FullFrame)
    {
      row_stride *= 2;
    }
    else
    {
      if (ftDecode == FrameAB)
      {
        outBuffer += dinfo.output_height*dinfo.output_width*dinfo.output_components;
      }
    }
    for (y = 0; y < dinfo.output_height; y++){
      jpeg_read_scanlines(&dinfo, (JSAMPARRAY) &outBuffer, 1);
      if (DoubleLines == TRUE)
      {
        memcpy(outBuffer+row_stride, outBuffer, row_stride);
        outBuffer += row_stride;
      }
      outBuffer += row_stride;
    }
    jpeg_destroy_decompress(&dinfo);
    
  }
  return 1;
}


/****************************************************************/
/* read the DIB to outBuffer with format YUV (3 byte per pixel) */
/****************************************************************/

int Decode_RGB(FILE *inFile, char *outBuffer){
uCHAR *Y;
uCHAR *U, *V;
int   y, u, v;
LONG  R, G, B;
WORD  *Pixel;
int   w, h;
uCHAR *row, *Crow;
int   row_stride;

  switch (GlobalInfo.vBitsPerPixel){
  case 1:
  case 4:
  case 8:
    fprintf(stderr, "color palet not supported\n");
    exit(1);
  case 16:
  case 24:
  case 32:
    w = GlobalInfo.vBitsPerPixel/8;
    row_stride = GlobalInfo.Width*w+((GlobalInfo.Width*w) % 4);
    row = malloc(row_stride);
    Y = outBuffer+(GlobalInfo.Height-1)*GlobalInfo.Width*3;
    U = Y+1;
    V = Y+2;
    for (h=GlobalInfo.Height-1; h>=0; h--){
      fread(row, row_stride, 1, inFile);
      Crow = row;
      Pixel = (WORD *)row;
      for (w=0; w<GlobalInfo.Width; w++){
        switch (GlobalInfo.vBitsPerPixel){
        case 16:
          R = (DWORD) (((*Pixel >> 7) & 0xf8) | 0x07);
          G = (DWORD) (((*Pixel >> 2) & 0xf8) | 0x07);
          B = (DWORD) (((*Pixel << 3) & 0xf8) | 0x07);
          Pixel++;
          break;
        case 24:
          B = (LONG)*Crow++;
          G = (LONG)*Crow++;
          R = (LONG)*Crow++;
          break;
        case 32:
          B = (LONG)*Crow++;
          G = (LONG)*Crow++;
          R = (LONG)*Crow++;
          Crow++;
          break;
        default:
          R=B=G=0; /* only to avoid warnings, is NEVER reached */
        }
        y = (G*(LONG)256+R*(LONG)130+B*(LONG)50)/438;
        u = B-y;
        v = R-y;
        *Y = (uCHAR)y;
        *U = (uCHAR)((u*127)/226+127);
        *V = (uCHAR)((v*127)/179+127);
        Y += 3;
        U += 3;
        V += 3;
      }
      Y -= GlobalInfo.Width*3*2;
      U -= GlobalInfo.Width*3*2;
      V -= GlobalInfo.Width*3*2;
    }
    free(row);
    break;
  default:
    fprintf(stderr, "unknown color depth\n");
    exit(1);
  }
  return 1;
}

/****************************************************************/
/* read the YVU to outBuffer with format YUV (3 byte per pixel) */
/****************************************************************/

int Decode_YVU9(FILE *inFile, char *outBuffer){
uCHAR *Y;
uCHAR *U, *V;
int   w, h, i, c;
uCHAR *row, *Crow;
int   row_stride;

  row_stride = GlobalInfo.Width;
  row = malloc(row_stride);
  Y = outBuffer;
  for (h=0; h<GlobalInfo.Height; h++){
    fread(row, row_stride, 1, inFile);
    Crow = row;
    for (w=0; w<GlobalInfo.Width; w++){
      *Y = *Crow++;
      Y += 3;
    }
  }
  row_stride >>= 2;
  V = outBuffer+2;
  for (h=0; h<(GlobalInfo.Height >> 2); h++){
    fread(row, row_stride, 1, inFile);
    for (i=0; i<4; i++){
      Crow = row;
      /* one V value for 4 pixel */
      for (w=0; w<row_stride; w++){
        c = *Crow++;
        *V = c;
        V += 3;
        *V = c;
        V += 3;
        *V = c;
        V += 3;
        *V = c;
        V += 3;
      }
    }
  }
  U = outBuffer+1;
  for (h=0; h<(GlobalInfo.Height >> 2); h++){
    fread(row, row_stride, 1, inFile);
    for (i=0; i<4; i++){
      Crow = row;
      /* one U value for 4 pixel */
      for (w=0; w<row_stride; w++){
        c = *Crow++;
        *U = c;
        U += 3;
        *U = c;
        U += 3;
        *U = c;
        U += 3;
        *U = c;
        U += 3;
      }
    }
  }
  free(row);
  return 1;
}


/****************************************************************/
/* read the Y41P to outBuffer with format YUV (3 byte per pixel) */
/****************************************************************/

int Decode_Y41P(FILE *inFile, char *outBuffer){
uCHAR *Y;
uCHAR *U, *V, u1, u2, v1, v2;
int   w, h, i;
uCHAR *row, *Crow;
int   row_stride;

  row_stride = GlobalInfo.Width+(GlobalInfo.Width >> 1);
  row = malloc(row_stride);
  Y = outBuffer;
  for (h=0; h<GlobalInfo.Height; h++){
    fread(row, row_stride, 1, inFile);
    Crow = row;
    Y = outBuffer+(3*GlobalInfo.Width)*(GlobalInfo.Height-(h+1));
    U = Y+1;
    V = Y+2;
    for (w=0; w<(GlobalInfo.Width >> 3); w++){
      u1 = *Crow++;
      *Y = *Crow++;
      Y += 3;
      v1 = *Crow++;
      *Y = *Crow++;
      Y += 3;
      u2 = *Crow++;
      *Y = *Crow++;
      Y += 3;
      v2 = *Crow++;
      *Y = *Crow++;
      Y += 3;
      for (i=0; i<4; i++){
        *Y = *Crow++;
        Y += 3;
      }
      for (i=0; i<4; i++){
        *U = u1;
        U += 3;
        *V = v1;
        V += 3;
      }
      for (i=0; i<4; i++){
        *U = u2;
        U += 3;
        *V = v2;
        V += 3;
      }
    }
  }
  free(row);
  return 1;
}


/******************************************************************/
/* scale and crop the YUV picture from PicSizeX x PicSizeY to     */
/* OutSizeX x OutSizeY and split to yBuffer, uBuffer and vBuffer. */
/* if OutSize > PicSize pictures is centered                      */
/******************************************************************/

void ScaleYUV(uCHAR *inBuffer, uCHAR* yBuffer, uCHAR *uBuffer, uCHAR *vBuffer,
              int PicSizeX, int PicSizeY, int OutSizeX, int OutSizeY,
              int CropLeft, int CropTop, int CropRight, int CropBottom){

  int ErrorX, ErrorY;
  int Scale1X,Scale2X, Scale1Y, Scale2Y;
  int ScaleSizeX, ScaleSizeY;
  int CountX, CountY, x, y;
  float Aspect;
  
  inBuffer += CropTop*3*PicSizeX;
  PicSizeX = PicSizeX-CropLeft-CropRight;
  PicSizeY = PicSizeY-CropTop-CropBottom;

  if (KeepAspect == TRUE){
    Aspect = (float)PicSizeX / (float)PicSizeY;
    if (((float)OutSizeX / (float)OutSizeY) > Aspect) {
      ScaleSizeY = OutSizeY;
      if (ScaleSizeY > PicSizeY) ScaleSizeY = PicSizeY;
      ScaleSizeX = (int)((float)ScaleSizeY * Aspect + 0.5555555);
      if (ScaleSizeX > OutSizeX) ScaleSizeX = OutSizeX;
    }
    else{
      ScaleSizeX = OutSizeX;
      if (ScaleSizeX > PicSizeX) ScaleSizeX = PicSizeX;
      ScaleSizeY = (int)((float)ScaleSizeX / Aspect + 0.5555555);
      if (ScaleSizeY > OutSizeY) ScaleSizeY = OutSizeY;
    }
  }
  else{
    ScaleSizeX = OutSizeX;
    ScaleSizeY = OutSizeY;
  }
  if (ScaleSizeX > PicSizeX) ScaleSizeX = PicSizeX;
  if (ScaleSizeY > PicSizeX) ScaleSizeY = PicSizeX;
  
  Scale1X = 2*ScaleSizeX;
  ErrorX = Scale1X-PicSizeX;
  Scale2X = ErrorX-PicSizeX;
  
  Scale1Y = 2*ScaleSizeY;
  ErrorY = Scale1Y-PicSizeY;
  Scale2Y = ErrorY-PicSizeY;
  
  CountY = 0;
  while (CountY < ((OutSizeY - ScaleSizeY) / 2)) {
    memset(yBuffer, 0, OutSizeX);
    memset(uBuffer, CENTERJSAMPLE, OutSizeX);
    memset(vBuffer, CENTERJSAMPLE, OutSizeX);
    yBuffer += OutSizeX;
    uBuffer += OutSizeX; 
    vBuffer += OutSizeX; 
    CountY++;
  }
  for (y=1; y<=PicSizeY; y++){
    inBuffer += 3*CropLeft;
    if (ErrorY >= 0){
      CountX = 0;
      while(CountX < ((OutSizeX-ScaleSizeX)/2)){
        *yBuffer++ = 0;
        *uBuffer++ = CENTERJSAMPLE; 
        *vBuffer++ = CENTERJSAMPLE; 
        CountX++;
      }
      ErrorX = Scale1X-PicSizeX;
      for (x=1; x<=PicSizeX; x++){
        if (ErrorX >= 0){
          *yBuffer++ = *inBuffer++;
          *uBuffer++ = *inBuffer++;
          *vBuffer++ = *inBuffer++;
          CountX++;
          ErrorX += Scale2X;
        } /* ScaleX */
        else{
          inBuffer += 3;
          ErrorX += Scale1X;
        } /* if ErrorX */
      } /* next x */
      while(CountX < OutSizeX){
        *yBuffer++ = 0;
        *uBuffer++ = CENTERJSAMPLE; 
        *vBuffer++ = CENTERJSAMPLE; 
        CountX++;
      }
      CountY++;
      ErrorY += Scale2Y;
    } /* if ErrorY */
    else{
      inBuffer += 3*PicSizeX;
      ErrorY += Scale1Y;
    } /* if ScaleY */
    inBuffer += 3*CropRight;
  } /* next y */
  while(CountY < OutSizeY){
    memset(yBuffer, 0, OutSizeX);
    memset(uBuffer, CENTERJSAMPLE, OutSizeX);
    memset(vBuffer, CENTERJSAMPLE, OutSizeX);
    yBuffer += OutSizeX;
    uBuffer += OutSizeX; 
    vBuffer += OutSizeX; 
    CountY++;
  }
}


/**************************************************************************/
/* Save the YUV picture in yBuffer, uBuffer and vBuffer to a UCB YUV file */
/* usable with mpeg_encode. U and V will be scaled half X and Y if Scale  */
/* is TRUE. If Name == NULL the output will go to stdout.                 */
/**************************************************************************/

void SaveYUV(char *Name, uCHAR *yBuffer, uCHAR *uBuffer, uCHAR *vBuffer,
             int PicSizeX, int PicSizeY, boolean Scale){
  FILE *outFile;
  WORD x, y;
  uCHAR *outU, *orgU, *outV, *orgV;

  if (Name == NULL)
    outFile = stdout;
  else
    if ((outFile = fopen(Name,"wb")) == NULL){
      fprintf(stderr, "could not open output file %s\n",Name);
      exit(1);
    }
  /* write Y */
  fwrite(yBuffer, PicSizeX*PicSizeY, 1, outFile);
  
  orgU = outU = uBuffer;
  orgV = outV = vBuffer;
  /* scale U, V */
  if (Scale){
    for (y=0; y<(PicSizeY >> 1); y++){
      for (x=0; x<(PicSizeX >> 1); x++){
        *outU++ = ((*uBuffer)+ (*(uBuffer+1))+
                  (*(uBuffer+PicSizeX))+ (*(uBuffer+PicSizeX+1))) / 4;
        *outV++ = ((*vBuffer)+ (*(vBuffer+1))+
                  (*(vBuffer+PicSizeX))+ (*(vBuffer+PicSizeX+1))) / 4;
        uBuffer += 2;
        vBuffer += 2;
      } 
      uBuffer += PicSizeX;
      vBuffer += PicSizeX;
    }
  }
  fwrite(orgU, PicSizeX*PicSizeY / 4, 1, outFile);
  fwrite(orgV, PicSizeX*PicSizeY / 4, 1, outFile);
  if (Name != NULL)
    fclose(outFile);
}


/**************************************************************************/
/* Save the YUV picture in yBuffer, uBuffer and vBuffer to a Windows BMP  */
/* file. If Name == NULL the output will go to stdout.                    */
/**************************************************************************/

void SaveBMP(char *Name, uCHAR *yBuffer, uCHAR *uBuffer, uCHAR *vBuffer,
             int PicSizeX, int PicSizeY){
  FILE  *outFile;
  WORD  x, y;
  uCHAR *line, *ptr;
  int   r,g,b, Y, U, V;
  WORD              BMP_ID;
  tBMPHeader        BMPHeader;
  tBitmapInfoHeader Header;

  if (Name == NULL)
    outFile = stdout;
  else
    if ((outFile = fopen(Name,"wb")) == NULL){
      fprintf(stderr, "could not open output file %s\n",Name);
      exit(1);
    }
  
  BMP_ID      = 0x4d42;
  fwrite(&BMP_ID, 2, 1, outFile);
  
  BMPHeader.bfSize      = PicSizeX*PicSizeY*3+sizeof(tBMPHeader)+
                            sizeof(tBitmapInfoHeader);
  BMPHeader.bfReserved1 = 0;
  BMPHeader.bfReserved2 = 0;
  BMPHeader.bfOffs      = sizeof(tBMPHeader)+sizeof(tBitmapInfoHeader)+2;

  fwrite(&BMPHeader, sizeof(tBMPHeader), 1, outFile);
  
  Header.biSize          = sizeof(tBitmapInfoHeader);
  Header.biWidth         = PicSizeX;
  Header.biHeight        = PicSizeY;
  Header.biPlanes        = 1;
  Header.biBitCount      = 24;
  Header.biCompression   = 0;
  Header.biSizeImage     = PicSizeX*PicSizeY*3;
  Header.biXPelsPerMeter = 96;
  Header.biYPelsPerMeter = 96;
  Header.biClrUsed       = 0;
  Header.biClrImportant  = 0;
  
  fwrite(&Header, sizeof(tBitmapInfoHeader), 1, outFile);
  
  line = malloc(PicSizeX*3);
  
  yBuffer += (PicSizeY-1) * PicSizeX;
  uBuffer += (PicSizeY-1) * PicSizeX;
  vBuffer += (PicSizeY-1) * PicSizeX;
  for (y=0; y<PicSizeY; y++){
    ptr = line;
    for (x=0; x<PicSizeX; x++){
      Y = *yBuffer++;
      U = (unsigned char)*uBuffer++;
      V = (unsigned char)*vBuffer++;
      U -= 128;
      V -= 128;
      if ((r = Y+V) < 0) r = 0;
      else
        if (r > 255) r = 255;
      if ((b = Y+U) < 0) b = 0;
      else
        if (b > 255) b = 255;
      g = ((Y*436L) >> 8) - ((r*130U) >> 8) - ((b*50U) >> 8);
      if (g < 0) g = 0;
      else
        if (g > 255) g = 255;
      *ptr++ = b;
      *ptr++ = g;
      *ptr++ = r;
    }
    fwrite(line, PicSizeX*3, 1, outFile);
    yBuffer -= 2*PicSizeX;
    uBuffer -= 2*PicSizeX;
    vBuffer -= 2*PicSizeX;
  }

  free(line);
    
  if (Name != NULL)
    fclose(outFile);
}


/***************************************************************************/
/* show the info of the AVI file                                           */
/***************************************************************************/

void PrintInfo(){
  char formstr[6];
  printf("Videostream info\n");
  printf("Width            : %i\n", GlobalInfo.Width);
  printf("Height           : %i\n", GlobalInfo.Height);
  printf("TotalFrames      : %li\n", GlobalInfo.Frames);
  printf("MicroSecPerFrame : %li (%2.3f fps)\n", GlobalInfo.MSPerFrame,
                             (float)1000000/(float)GlobalInfo.MSPerFrame);
  printf("Scale/Rate       : %li/%li\n", GlobalInfo.Scale, GlobalInfo.Rate);
  FOURCC2Str(GlobalInfo.vHandler, formstr);
  printf("Video Handler    : %s\n", formstr);
  printf("Planes           : %i\n", GlobalInfo.vPlanes);
  printf("Bits per pixel   : %i\n", GlobalInfo.vBitsPerPixel);
  if (GlobalInfo.vCompression > 2){
    FOURCC2Str(GlobalInfo.vCompression, formstr);
    printf("Compression      : %s\n", formstr);
  }
  else {
    printf("Compression      : ");
    switch (GlobalInfo.vCompression){
    case 0: printf("RGB\n");
            break;
    case 1: printf("RLE8\n");
            break;
    case 2: printf("RLE4\n");
            break;
    }
  }
  printf("Audiostream info\n");
  switch (GlobalInfo.aFormat){
  case 1: printf("Wave format      : PCM\n");
          break;
  case 2: printf("Wave format      : ADPCM\n");
          break;
  default:printf("Wave format      : unknown\n");
  }
  printf("Channels         : %i\n", GlobalInfo.aChannels);
  printf("SamplesPerSec    : %li\n", GlobalInfo.aSamplesPerSec);
  printf("BitsPerSample    : %i\n", GlobalInfo.aBitsPerSample);
}

/***************************************************************************/
/* Processing of a chunk (will be called recursively!)                     */
/***************************************************************************/

boolean ProcessChunk(FILE* f, long filepos, long filesize,
                     FOURCC DesiredTag, DWORD* chunksize)
{

  tMainAVIHeader    MainAVIHeader;
  tAVIStreamHeader  AVIStreamHeader;
  tBitmapInfoHeader BitmapInfoHeader;
  tWaveFormatEx     WaveFormatEx;
  
  char   tagstr[5];          /* FOURCC of chunk converted to string */
  FOURCC chunkid;            /* read FOURCC of chunk                */
  long   datapos;            /* position of data in file to process */
  int    x, y;
  long   soundsize;

  if (filepos>filesize-1) {  /* Oops. Must be something wrong!      */
    if (PrintFormat)
      printf("\n\n *** Error: Data would be behind end of file!\n");
    GlobalInfo.vError = 1;
    return(FALSE);
  }
  fseek(f,filepos,SEEK_SET);    /* Go to desired file position!     */

  if (!ReadChunkHead(f,&chunkid,chunksize)) {  /* read chunk header */
    if (PrintFormat)
      printf("\n\n *** Error reading chunk at filepos 0x%lx\n",filepos);
    GlobalInfo.vError = 2;
    return(FALSE);
  }
  if (DesiredTag) {                 /* do we have to test identity? */
    if (DesiredTag!=chunkid) {
      if (PrintFormat){
        char ds[5];
        FOURCC2Str(DesiredTag,ds);
        printf("\n\n *** Error: Expected chunk '%s', found '%s'\n",
               ds,tagstr);
      }
      GlobalInfo.vError = 3;
      return(FALSE);
    }
  }

  datapos=filepos+sizeof(FOURCC)+sizeof(DWORD); /* here is the data */

  if (datapos + ((*chunksize+1)&~1) > filesize) {      /* too long? */
    if ((chunkid != RIFFtag) && (chunkid != LISTtag))
    {
      if (PrintFormat){
        printf("\n *** Error: Chunk exceeds file (starts at 0x%lx) ***\n",
               filepos);
      }
      GlobalInfo.vError = 4;
      return(FALSE);
    }
  }

  switch (chunkid) {

  /* Depending on the ID of the chunk and the internal state, the
     different IDs can be interpreted. At the moment the only
     interpreted chunks are RIFF- and LIST-chunks. For all other
     chunks only their header is printed out. */

    case RIFFtag:
    case LISTtag: {

      DWORD datashowed;
      FOURCC formtype;       /* format of chunk                     */
      DWORD  subchunksize;   /* size of a read subchunk             */

      LastChunkTyp = ChunkLIST;
      
      fread(&formtype,sizeof(FOURCC),1,f);    /* read the form type */

      datashowed=sizeof(FOURCC);    /* we showed the form type      */
      datapos+=datashowed;          /* for the rest of the routine  */
      
      if (formtype == movitag){     /* the header is parsed         */
        DWORD size;
        
        if (CropX2 == -1) CropX2 = 0;
        if (CropY2 == -1) CropY2 = 0;
        if (FrameSizeX == -1) FrameSizeX = GlobalInfo.Width-CropX1-CropX2;
        if (FrameSizeY == -1) FrameSizeY = GlobalInfo.Height-CropY1-CropY2;

        if (LastFrame == -1) LastFrame = GlobalInfo.Frames;
        if (MSPerFrame == 0){
          MSPerFrame = GlobalInfo.MSPerFrame;
        }
        if (MSPerFrame > GlobalInfo.MSPerFrame){
          fprintf(stderr, "new frame rate is to low\n");
          exit(1);
        }
        if (MSPerFrame != GlobalInfo.MSPerFrame){
          LastFrame = (int) ((DWORD)LastFrame * GlobalInfo.MSPerFrame / MSPerFrame);
        }
        if (GlobalInfo.aBitsPerSample == 8) {
          soundsize = GlobalInfo.aChannels*
                      ((GlobalInfo.aSamplesPerSec * MSPerFrame) / 1000000);
        }
        else {
          soundsize = GlobalInfo.aChannels*2*
                      ((GlobalInfo.aSamplesPerSec * MSPerFrame) / 1000000);
        }
        SoundStart = soundsize*(FirstFrame-1);
        SoundEnd   = soundsize*(LastFrame-1);
        SoundPos = 0;
        
        // don't allocate memory if we only want to know the format
        if (PrintFormat){
          PrintInfo();
          GlobalInfo.vError = 0;
          if (!TestFrames)
            return FALSE;
        }
        
        x = GlobalInfo.Width;
        y = GlobalInfo.Height;
        if (FrameSizeX > GlobalInfo.Width) x = FrameSizeX;
        if (FrameSizeY > GlobalInfo.Height) y = FrameSizeY;
        size = x*y;
        outBuffer = (uCHAR*) malloc(size*3);
        YBuffer = (uCHAR*) malloc(size);
        UBuffer = (uCHAR*) malloc(size);
        VBuffer = (uCHAR*) malloc(size);
      }

      while (datashowed<*chunksize) {      /* while not showed all: */

        long subchunklen;           /* complete size of a subchunk  */

        /* recurse for subchunks of RIFF and LIST chunks: */
        if (!ProcessChunk(f, datapos, filesize, 0,
                          &subchunksize))
        {
          GlobalInfo.vError = 0;
          return(FALSE);
        }

        subchunklen = sizeof(FOURCC) +  /* this is the complete..   */
                      sizeof(DWORD)  +  /* .. size of the subchunk  */
                      ((subchunksize+1) & ~1);

        datashowed += subchunklen;      /* we showed the subchunk   */
        datapos    += subchunklen;      /* for the rest of the loop */
      }
    } break;
    /* RIFF and LIST end */
    case avihtag: {
      fread(&MainAVIHeader, sizeof(tMainAVIHeader),1,f);
      GlobalInfo.Width = MainAVIHeader.dwWidth;
      GlobalInfo.Height = MainAVIHeader.dwHeight;
      GlobalInfo.Frames = MainAVIHeader.dwTotalFrames;
      GlobalInfo.MSPerFrame = MainAVIHeader.dwMicroSecPerFrame;
      LastChunkTyp = ChunkHeader;
    } break;
    /* AVIH end */
    case strhtag: {
      char formstr[5];
      fread(&AVIStreamHeader, sizeof(tAVIStreamHeader),1,f);
      FOURCC2Str(AVIStreamHeader.fccType, formstr);
      if (formstr[0] == 'v'){
        LastChunkTyp = ChunkVIDS;
        GlobalInfo.vHandler = AVIStreamHeader.fccHandler;
        GlobalInfo.Scale = AVIStreamHeader.dwScale;
        GlobalInfo.Rate = AVIStreamHeader.dwRate;
      }
      else {
        LastChunkTyp = ChunkAUDS;
      }
    } break;
    /* STRH end */
    case strftag: {
      switch (LastChunkTyp){
        case ChunkVIDS: {
          fread(&BitmapInfoHeader, sizeof(tBitmapInfoHeader),1,f);
          GlobalInfo.vPlanes = BitmapInfoHeader.biPlanes;
          GlobalInfo.vBitsPerPixel = BitmapInfoHeader.biBitCount;
          GlobalInfo.vCompression = BitmapInfoHeader.biCompression;
          if ((GlobalInfo.vCompression == 0) && (GlobalInfo.vHandler == 0))
            GlobalInfo.vHandler = TagDIB;
        } break;
        /* VIDS end */
        case ChunkAUDS:{
          fread(&WaveFormatEx, sizeof(tWaveFormatEx),1,f);
          GlobalInfo.aFormat = WaveFormatEx.wFormatTag;
          GlobalInfo.aChannels = WaveFormatEx.nChannels;
          GlobalInfo.aSamplesPerSec = WaveFormatEx.nSamplesPerSec;
          GlobalInfo.aBitsPerSample = WaveFormatEx.wBitsPerSample;
        } break;
        /* AUDS end */
      } /* switch end */
    } break;
    /* STRF end */
    case Tag00db:
    case Tag00dc: /* for Adobe Premiere 4.0 */ {
      PicNumber++;
      if ((PicNumber > LastFrame) && (WaveFile == NULL)){
        if (SaveFrames) printf("%li frames saved\n", PicNumber-1);
        exit(0);
      }
      if ((PicNumber < FirstFrame) || (PicNumber > LastFrame))
      {
        break;
      }
      if (!PrintFormat)
      {
        char Name[255];
        if ((ftSave == FrameA) && (CountSkip != 0))
        {
          sprintf(Name, PicName, PicNumber*2);
        }
        else
        {
          if ((ftSave == FrameB) && (CountSkip != 0))
          {
            sprintf(Name, PicName, PicNumber*2+CountSkip);
          }
          else
          {
            sprintf(Name, PicName, PicNumber);
          }
        }
        if (SaveFrames)
        {
          printf("saving frame %li\r", PicNumber);
          fflush(stdout);
        }
        if ((*chunksize) != 0)
        {
          int DecodeRes;
          switch (GlobalInfo.vHandler){
          case TagMJPG:
          case Tagmjpg:
          case Tagdmb1:
            {
              if (GlobalInfo.Height < 289)
              // so we don't have interlaced video
              // could only be shrinked NTSC which shouldn't happen
              {
                DecodeRes = Decode_JPEG(f, outBuffer);
              }
              else
              {
                DecodeRes = Decode_JPEG_interlaced(f, outBuffer, ftSave, DoubleLines);
              }
            }
            break;
          case TagDIB:
            if (GlobalInfo.vCompression != 0){
              fprintf(stderr, "unsupported video compression\n");
              exit(1);
            }
            DecodeRes = Decode_RGB(f, outBuffer);
            break;
          case TagYVU9:
            DecodeRes = Decode_YVU9(f, outBuffer);
            break;
          case TagY41P:
            DecodeRes = Decode_Y41P(f, outBuffer);
            break;
          default:
            fprintf(stderr, "unsupported video handler\n");
            exit(1);
          }
          if (DecodeRes == 1)
          {
            /* if a new frame could be decoded without error scale it.
               If an error happened while decoding output the previous frame
               (hoping we could decode the first frame without errors)      */
            ScaleYUV(outBuffer, YBuffer, UBuffer, VBuffer, GlobalInfo.Width, 
                     GlobalInfo.Height, FrameSizeX, FrameSizeY, CropX1, CropY1,
                     CropX2, CropY2);
          }
          if (SaveFrames)
          {
            if (PicFormat == 0)
              SaveYUV(Name, YBuffer, UBuffer, VBuffer, FrameSizeX, FrameSizeY, TRUE);
            else
              SaveBMP(Name, YBuffer, UBuffer, VBuffer, FrameSizeX, FrameSizeY);
          }
          else {
            SaveYUV(NULL, YBuffer, UBuffer, VBuffer, FrameSizeX, FrameSizeY, TRUE);
          }
          // we've decoded a frame so increse the timer;
          AVI_MS += GlobalInfo.MSPerFrame;
          ActualMS += MSPerFrame;
        } // chunksize != 0
        else
        {
          DroppedFrames++;
        }
        while (((ActualMS+MSPerFrame) < AVI_MS) &&
               (PicNumber <= LastFrame)){
          /* repeat last frame to get a higher frame rate */
          PicNumber++;
          if (SaveFrames){
            char Name[255];
            sprintf(Name, PicName, PicNumber);
            printf("saving frame %li\r", PicNumber);
            fflush(stdout);
            SaveYUV(Name, YBuffer, UBuffer, VBuffer, FrameSizeX, FrameSizeY, FALSE);
          }
          else {
            SaveYUV(NULL, YBuffer, UBuffer, VBuffer, FrameSizeX, FrameSizeY,
                    FALSE);
          }
          ActualMS += MSPerFrame;
        } /* while */
      } /* !PrintFormat */
      else
      {
        if (TestFrames == TRUE)
        {
          switch (GlobalInfo.vHandler)
          {
            case TagMJPG:
            case Tagmjpg:
            case Tagdmb1:
              { unsigned char buf[4];
                fread(buf, 4,1,f);
                if ((buf[0] != 0xff) || (buf[1] != 0xd8) ||
                    (buf[2] != 0xff) || (buf[3] != 0xe0)) {
                  CorruptFrames++;
                  fprintf(stderr, "- Corrupt frame %li\n", PicNumber);
                }
              }
              break;
            default:
              /* is there any other format to test (e.g BtYUV)? */
              break;
          }
        }
      } // PrintFormat
    } break;
    /* picture end */
    case Tag01wb: {
      char *buffer;
      if (WaveFile != NULL){
        if ((SoundPos+ *chunksize) >= SoundStart) {
          if (SoundPos < SoundEnd) {
            buffer = malloc(*chunksize);
            fread(buffer, *chunksize, 1, f);
            if (SoundPos < SoundStart) {
              long size;
              size = *chunksize-(SoundStart-SoundPos);
              fwrite(&buffer[SoundStart-SoundPos], size, 1, WaveFile);
            }
            else
              fwrite(buffer, *chunksize, 1, WaveFile);
            free(buffer);
          }
        }
        SoundPos += *chunksize;
      }
    } break;
    /* WAVE end */
  }

  GlobalInfo.vError = 0;
  return(TRUE);
}


/**********************/
/* print a short help */
/**********************/

void Usage(char *Name){
  printf("%s version 0.97\n", Name);
  printf("Convert AVIs (MJPEG and RGB) to YUV and output to STDOUT\n");
  printf("Usage: %s [option] filename\n",Name); 
  printf("[option] = -aNNN           first frame is NNN\n");
  printf("           -cbCCC          crop CCC points at the bottom of the picture\n");
  printf("           -clCCC          crop CCC points at the left side of the picture\n");
  printf("           -crCCC          crop CCC points at the right side of the picture\n");
  printf("           -ctCCC          crop CCC points at the top of the picture\n");
  printf("                           cropping is done before scaling.\n");
  printf("           -eNNN           last frame is NNN\n");
  printf("           -fFPS           set the frame rate of the output stream to FPS\n");
  printf("                           FPS mast be >= FPS of the AVI\n");
  printf("           -sFrame%%i.yuv   save each frame as file. The name must contain\n");
  printf("                           one 'printf' integer format. .yuv or .bmp\n");
  printf("           -v              print the format of the AVI\n");
  printf("           -t              test the correctness of MJPEG frames\n");
  printf("           -xAAA           scale the output pictures to width AAA\n");
  printf("           -yBBB           scale the output pictures to height BBB\n");
  printf("           -z              don't keep the aspect ration when scaling\n");
  printf("           -wWavefile      name of the file the audio is written to\n");
  printf("                           Formats are .pcm or .wav\n");
  printf("           -nX             X = a  save field A\n");
  printf("                               b  save field B\n");
  printf("                               A  save field A and double every line\n");
  printf("                               B  save field B and double every line\n");
  printf("                               f  save field A+B interleaved\n");
  printf("                               m  save field A+B in one picture\n");
  printf("           -i+/-           save field A with even and field B with odd numbers\n");
  printf("                           i+  files B follows field A\n");
  printf("                           i-  files A follows field B\n");
  printf("           filename        name of the AVI\n\n");
}


/******************************************/
/* main routine with command line parsing */
/******************************************/

int main (int argc, char **argv)
{
  FILE*       f;            /* the input file              */
  long        filesize;     /* its size                    */
  DWORD       chunksize;    /* size of the RIFF chunk data */
  int         i;
  int         NameIndex;    /* commandline index of the AVI-filename */
  float       Aspect, ScaleX, ScaleY;
  int         ScaleSizeX, ScaleSizeY;

  if (argc<2) { 
    Usage(argv[0]);
    return(1);
  }

  PrintFormat = FALSE;
  FrameSizeX = -1;
  FrameSizeY = -1;
  f = NULL;
  WaveFile = NULL;
  NameIndex = 1;
  SaveFrames = FALSE;
  KeepAspect = TRUE;
  ActualMS = 0;
  AVI_MS = 0;
  TestFrames = FALSE;
  ftSave = FrameA;

  for (i=1; i<argc; i++){
    if (*argv[i] == '-'){
      switch (argv[i][1]){
        case 'v': PrintFormat = TRUE;
                  break;
        case 't': TestFrames = TRUE;
                  PrintFormat = TRUE; /* doesn't work without */
                  break;
        case 'x': FrameSizeX = atoi(&(argv[i][2]));
                  if (FrameSizeX == 0){
                    fprintf(stderr, "frame size X missing\n");
                    exit(1);
                  }
                  break;
        case 'y': FrameSizeY = atoi(&(argv[i][2]));
                  if (FrameSizeY == 0){
                    fprintf(stderr, "frame size Y missing\n");
                    exit(1);
                  }
                  break;
        case 'w': if ((WaveFile = fopen(&(argv[i][2]), "wb")) == NULL){
                    fprintf(stderr, "can't open wave file %s\n", &(argv[i][2]));
                    exit(1);
                  }
                  if (argv[i][strlen(argv[i])-1] == 'v') {
                    memset(&WaveHeader, 0, sizeof(WaveHeader));
                    WaveFormat = 1;
                    WaveHeader.fccMainChunk = RIFFtag;
                    WaveHeader.fccType      = TagWAVE;
                    WaveHeader.fccSubType   = Tagfmt;
                    WaveHeader.dwLength     = 16;
                    WaveHeader.wFormat      = 1;
                    WaveHeader.fccData      = Tagdata;
                    fwrite(&WaveHeader, sizeof(WaveHeader), 1, WaveFile);
                  }
                  break;
        case 'a': FirstFrame = atoi(&(argv[i][2]));
                  if (FirstFrame <= 0){
                    fprintf(stderr, "first frame must be >= 1\n");
                    exit(1);
                  }
                  break;
        case 'e': LastFrame = atoi(&(argv[i][2]));
                  if (LastFrame < FirstFrame) {
                    fprintf(stderr, "last frame must be >= first frame\n");
                    exit(1);
                  }
                  break;
        case 's': SaveFrames = TRUE;
                  if (strlen(argv[i]) > 255){
                    fprintf(stderr, "length of filename must be less than 255\n");
                    exit(1);
                  }
                  strcpy(PicName, &(argv[i][2]));
                  if (PicName[strlen(PicName)-1] == 'p'){
                    printf("saving .BMP\n");
                    PicFormat = 1;
                  }
                  else {
                    printf("saving .YUV\n");
                    PicFormat = 0;
                  }
                  break;
        case 'f': MSPerFrame = (DWORD)(1000000/atof(&(argv[i][2])));
                  break;
        case 'c': {
                    switch (argv[i][2]){
                    case 'b': CropY2 = atoi(&(argv[i][3]));
                              if (CropY2 < 0) {
                                fprintf(stderr, "can't crop negative\n");
                                exit(1);
                              }
                              break;
                    case 'l': CropX1 = atoi(&(argv[i][3]));
                              if (CropX1 < 0) {
                                fprintf(stderr, "can't crop negative\n");
                                exit(1);
                              }
                              break;
                    case 'r': CropX2 = atoi(&(argv[i][3]));
                              if (CropX2 < 0) {
                                fprintf(stderr, "can't crop negative\n");
                                exit(1);
                              }
                              break;
                    case 't': CropY1 = atoi(&(argv[i][3]));
                              if (CropY1 < 0) {
                                fprintf(stderr, "can't crop negative\n");
                                exit(1);
                              }
                              break;
                    default:  fprintf(stderr, "unknown cropping\n");
                              exit(1);
                    }
                  }
                  break;
        case 'z': KeepAspect = FALSE;
                  break;
        case 'n': {
                    switch(argv[i][2]){
                    case 'a': ftSave = FrameA;
                              break;
                    case 'b': ftSave = FrameB;
                              break;
                    case 'A': ftSave = FrameA;
                              DoubleLines = TRUE;
                              break;
                    case 'B': ftSave = FrameB;
                              DoubleLines = TRUE;
                              break;
                    case 'f': ftSave = FullFrame;
                              break;
                    case 'm': ftSave = BothFrame;
                    default:  ftSave = FrameA;
                              break;
                    }
                  }
                  break;
        case 'i': {
                    if (argv[i][2] == '-')
                    {
                      CountSkip = -1;
                    }
                    else
                    {
                      CountSkip = 1;
                    }
                  }
                  break;
        default:  fprintf(stderr, "unknown option %s\n\n", argv[i]);
                  Usage(argv[0]);
                  exit(1);
      }
    }
    else{
      if (f == NULL){
        if (!(f=fopen(argv[i],"rb"))) {
          fprintf(stderr, "*** Error opening file %s. Program aborted!\n",
                  argv[i]);
          exit(1);
        }
        NameIndex = i;
      }
      else {
        fprintf(stderr, "unknown option %s\n\n", argv[i]);
        Usage(argv[0]);
        exit(1);
      }
    } /* *argv[i] != '-' */
  } /* next i */

  fseek(f, 0, SEEK_END);
  filesize = ftell(f);
  fseek(f, 0, SEEK_SET);

  if (PrintFormat)
    printf("Format of file %s (%ld bytes):\n\n",argv[NameIndex],filesize);

  /* Process the main chunk, which MUST be a RIFF chunk. */
  if (!ProcessChunk(f, 0, filesize, RIFFtag, &chunksize))
  {
    if (PrintFormat && !TestFrames && (GlobalInfo.vError != 0))
    {
      PrintFormat = 0;
      printf("\nerror %i while scanning header\n", GlobalInfo.vError);
      printf("use '-t' to get more info about the AVI\n");
    }
  }

  if (PrintFormat && TestFrames){
    i = FrameSizeY;
    printf("Info for mpeg_encode\n");
    if (KeepAspect == TRUE){
      Aspect = (float)GlobalInfo.Width / (float)GlobalInfo.Height;
      ScaleSizeX = FrameSizeX - CropX1 - CropX2;
      ScaleSizeY = FrameSizeY - CropY1 - CropY2;
      ScaleX = (float)FrameSizeX / (float)ScaleSizeX;
      ScaleY = (float)FrameSizeY / (float)ScaleSizeY;
      if (ScaleY < ScaleX) ScaleX = ScaleY;
      if (((float)ScaleSizeX / (float)ScaleSizeY) > Aspect) {
        ScaleSizeX = (int)((float)ScaleSizeX * Aspect * ScaleX + 0.5555555);
        ScaleSizeY = (int)((float)ScaleSizeY * ScaleX);
        if (ScaleSizeX > FrameSizeX) ScaleSizeX = FrameSizeX;
      }
      else{
        ScaleSizeX = (int)((float)ScaleSizeX * ScaleX);
        ScaleSizeY = (int)(((float)ScaleSizeX / Aspect)*ScaleX + 0.5555555);
        if (ScaleSizeY > FrameSizeY) ScaleSizeY = FrameSizeY;
      }
      printf("Picture size X   : %i (effective %i)\n", FrameSizeX, ScaleSizeX);
      printf("Picture size Y   : %i (effective %i)\n", FrameSizeY, ScaleSizeY);
    }
    else{
      printf("Picture size X   : %i\n", FrameSizeX);
      printf("Picture size Y   : %i\n", FrameSizeY);
    }
    printf("MicroSecPerFrame : %li (%2.3f fp/s)\n", MSPerFrame,
                               (float)1000000/(float)MSPerFrame);
    if (TestFrames) {
      printf("Frames scanned   : %li\n", PicNumber);
      printf("Dropped frames   : %li\n", DroppedFrames);
      printf("Corrupt frames   : %li\n", CorruptFrames);
    }
    printf("First frame      : %i\n", FirstFrame);
    printf("Last frame       : %li\n", PicNumber);
  }
  
  /*
  if (filesize != sizeof(FOURCC)+sizeof(DWORD)+ ((chunksize+1)&~1)){
    if (PrintFormat)
      printf("\n *** Warning: Padding bytes after RIFF chunk! ***\n");
  }
  */

  if (WaveFile != NULL) {
    if (WaveFormat == 1) {
      fflush(WaveFile);
      WaveHeader.dwFileLength    = ftell(WaveFile)-8;
      WaveHeader.wChannels       = GlobalInfo.aChannels;
      WaveHeader.dwSamplesPerSec = GlobalInfo.aSamplesPerSec;
      if (GlobalInfo.aBitsPerSample == 16)
        WaveHeader.wBytesPerSample = 2;
      else
        WaveHeader.wBytesPerSample = 1;
      WaveHeader.dwBytesPerSec = WaveHeader.dwSamplesPerSec *
                                   WaveHeader.wBytesPerSample;
      WaveHeader.wBitsPerSample  = GlobalInfo.aBitsPerSample;
      WaveHeader.dwDataLength    = ftell(WaveFile)-sizeof(WaveHeader);
    }
    fseek(WaveFile, 0, SEEK_SET);
    fwrite(&WaveHeader, sizeof(WaveHeader), 1, WaveFile);
    fclose(WaveFile);
  }
  if (SaveFrames) printf("%li frames saved\n", PicNumber);
  free(outBuffer);
  free(YBuffer);
  free(UBuffer);
  free(VBuffer);
  return(0);
}
